@@ -1,7 +1,7 @@
-__version__ = "1.2.0"
+__version__ = "1.7.1"
 
 import logging
-
+from urllib.parse import urlparse, parse_qs
 import dateutil.parser
 import requests
 
@@ -19,8 +19,10 @@ class QRonosClient(object):
     urls = {}
     headers = {}
 
-    def __init__(self, host, token=None):
+    def __init__(self, host, token=None, verify=True):
+        host = urlparse(host, scheme='https').geturl().rstrip('/')
         self._set_urls(host)
+        self.verify = verify
         if token:
             self._set_headers(token)
 
@@ -28,7 +30,7 @@ class QRonosClient(object):
         self.headers = {"Authorization": f"Token {token}"}
 
     def _set_urls(self, host):
-        prefix = f"https://{host}/api"  # TODO: Maybe use urllib.parse.urlparse to be more forgiving with host formats?
+        prefix = f"{host}/api"
         self.urls = {
             'login': f"{prefix}/login/",
             'logout': f"{prefix}/logout/",
@@ -46,7 +48,7 @@ class QRonosClient(object):
         if self.headers:
             kwargs.setdefault("headers", {}).update(self.headers)
         try:
-            return requests.get(*args, **kwargs)
+            return requests.get(*args, verify=self.verify, **kwargs)
         except requests.exceptions.RequestException as err:
             raise QRonosError("Unable to connect") from err
 
@@ -54,7 +56,7 @@ class QRonosClient(object):
         if self.headers:
             kwargs.setdefault("headers", {}).update(self.headers)
         try:
-            return requests.post(*args, **kwargs)
+            return requests.post(*args, verify=self.verify, **kwargs)
         except requests.exceptions.RequestException as err:
             raise QRonosError("Unable to connect") from err
 
@@ -90,7 +92,7 @@ class QRonosClient(object):
         elif response.status_code == 404:
             raise QRonosError(response.json()["job_id"])
         else:
-            raise QRonosError(f"Bad Request - {response.json()}")
+            raise QRonosError(f"Bad Request - {response.status_code} - {response.reason}")
 
     def _run_import(self, url, post_data):
         """Helper function to post import and handle response"""
@@ -164,7 +166,16 @@ class QRonosClient(object):
             }
         )
 
-    def get_item_attributes(self, show_mastered=True, show_non_mastered=True, tracker=None, unique_keys=None, unique_key=None):
+    def get_item_attributes(
+        self, 
+        show_mastered=True, 
+        show_non_mastered=True, 
+        tracker=None, 
+        unique_keys=None, 
+        unique_key=None,
+        page_size=None,
+        page=None,
+    ):
         """
         Get item attributes with a tracker ID or using a unique key(s)
 
@@ -174,6 +185,8 @@ class QRonosClient(object):
             :param tracker: The ID of the Tracker
             :param unique_keys: A list of unique item keys
             :param unique_key: A unique item key 
+            :param page_size: Number of items per page
+            :param page: The link for next page
         :return: Item attributes
         """
 
@@ -181,6 +194,7 @@ class QRonosClient(object):
             "show_mastered": show_mastered,
             "show_non_mastered": show_non_mastered,
         }
+        params = {}
 
         if tracker:
             post_data["tracker"] = tracker
@@ -188,15 +202,19 @@ class QRonosClient(object):
             post_data["unique_keys"] = unique_keys
         elif unique_key:
             post_data["unique_key"] = unique_key
+        if page:
+            params.update(parse_qs(urlparse(page).query))
+        if page_size:
+            params['page_size'] = page_size
 
-        response = self._post(self.urls['get_attributes'], json=post_data)
+        response = self._post(self.urls['get_attributes'], json=post_data, params=params)
         if response.status_code == 200:
             try:
-                return response.json()["items"]
+                return response.json()
             except Exception as err:
                 raise QRonosError("Unable to get attribute data") from err
-        elif response.status_code == 400:
-            raise QRonosError(f"Bad Request - {response.json()}")
+        else:
+            raise QRonosError(f"QRONOS Error - {response.status_code} - {response.reason}")
 
     def get_item_stages(
         self, 
@@ -207,6 +225,8 @@ class QRonosClient(object):
         stage=None,
         tracker_stage=None,
         tracker_stages=None,
+        page_size=None,
+        page=None,
     ):
         """
         Get item stages with a tracker ID or using a unique key(s) which can be restricted by stage(s)
@@ -219,10 +239,13 @@ class QRonosClient(object):
             :param stage: The ID of a stage (Stage model)
             :param tracker_stage: The ID of a tracker stage (TrackerStage model)
             :param tracker_stages: A list of tracker stage ids (TrackerStage model)
+            :param page_size: Number of items per page
+            :param page: The link for next page
         :return: Item stages
         """
 
         post_data = {}
+        params = {}
 
         if tracker:
             post_data["tracker"] = tracker
@@ -240,14 +263,19 @@ class QRonosClient(object):
         elif tracker_stages:
             post_data["tracker_stages"] = tracker_stages
 
-        response = self._post(self.urls['get_stages'], json=post_data)
+        if page:
+            params.update(parse_qs(urlparse(page).query))
+        if page_size:
+            params['page_size'] = page_size
+
+        response = self._post(self.urls['get_stages'], json=post_data, params=params)
         if response.status_code == 200:
             try:
-                return response.json()["items"]
+                return response.json()
             except Exception as err:
                 raise QRonosError("Unable to get stage data") from err
-        elif response.status_code == 400:
-            raise QRonosError(f"Bad Request - {response.json()}")
+        else:
+            raise QRonosError(f"QRONOS Error - {response.status_code} - {response.reason}")
 
     def get_item_stage_history(
             self,
@@ -258,6 +286,8 @@ class QRonosClient(object):
             stage=None,
             tracker_stage=None,
             tracker_stages=None,
+            page_size=None,
+            page=None,
     ):
 
         """
@@ -271,10 +301,13 @@ class QRonosClient(object):
             :param stage: The ID of a stage (Stage model)
             :param tracker_stage: The ID of a tracker stage (TrackerStage model)
             :param tracker_stages: A list of tracker stage ids (TrackerStage model)
+            :param page_size: Number of items per page
+            :param page: The link for next page
         :return: Item stages history
         """
 
         post_data = {}
+        params = {}
 
         if tracker:
             post_data["tracker"] = tracker
@@ -292,11 +325,85 @@ class QRonosClient(object):
         elif tracker_stages:
             post_data["tracker_stages"] = tracker_stages
 
-        response = self._post(self.urls['get_stage_history'], json=post_data)
+        if page:
+            params.update(parse_qs(urlparse(page).query))
+        if page_size:
+            params['page_size'] = page_size
+
+        response = self._post(self.urls['get_stage_history'], json=post_data, params=params)
         if response.status_code == 200:
             try:
                 return response.json()
             except Exception as err:
                 raise QRonosError("Unable to get stage history data") from err
-        elif response.status_code == 400:
-            raise QRonosError(f"Bad Request - {response.json()}")
+        else:
+            raise QRonosError(f"QRONOS Error - {response.status_code} - {response.reason}")
+
+    def get_all_item_attributes(self, **kwargs):
+        """
+        Get all item attributes with a tracker ID or using a unique key(s)
+
+        Keyword arguments:
+            :param show_mastered: Include mastered attributes
+            :param show_non_mastered: Include non-mastered attributes
+            :param tracker: The ID of the Tracker
+            :param unique_keys: A list of unique item keys
+            :param unique_key: A unique item key
+            :param page_size: Number of items per iteration
+        :return: Item attributes
+        """
+        return self._get_all_item_data(get_data_method=self.get_item_attributes, **kwargs)
+
+    def get_all_item_stages(self, **kwargs):
+        """
+        Get all item stages with a tracker ID or using a unique key(s) which can be restricted by stage(s)
+
+        Keyword arguments:
+            :param tracker: The ID of the Tracker
+            :param unique_keys: A list of unique item keys
+            :param unique_key: A unique item key
+            :param stages: A list of stage ids (Stage model)
+            :param stage: The ID of a stage (Stage model)
+            :param tracker_stage: The ID of a tracker stage (TrackerStage model)
+            :param tracker_stages: A list of tracker stage ids (TrackerStage model)
+            :param page_size: Number of items retrieved per iteration
+        :return: Item stages
+        """
+        return self._get_all_item_data(get_data_method=self.get_item_stages, **kwargs)
+
+    def get_all_item_stage_history(self, **kwargs):
+        """
+        Get all item stages history with a tracker ID or using a unique key(s) which can be restricted by stage(s)
+
+        Keyword arguments:
+            :param tracker: The ID of the Tracker
+            :param unique_keys: A list of unique item keys
+            :param unique_key: A unique item key
+            :param stages: A list of stage ids (Stage model)
+            :param stage: The ID of a stage (Stage model)
+            :param tracker_stage: The ID of a tracker stage (TrackerStage model)
+            :param tracker_stages: A list of tracker stage ids (TrackerStage model)
+            :param page_size: Number of retrieved per iteration
+        :return: Item stages history
+        """
+        return self._get_all_item_data(get_data_method=self.get_item_stage_history, **kwargs)
+
+    @staticmethod
+    def _get_all_item_data(get_data_method, **kwargs):
+        """
+        loops through and aggregates all pages for a given qronos client get_data method given the appropriate kwargs
+        Keyword arguments:
+            :param get_data_method: A qronos client get_data method
+        :return: all pages for given get method
+        """
+
+        # Need to remove page from kwargs so that it doesn't overwrite when calling get_data_method
+        kwargs.pop("page", None)
+
+        all_data = []
+        data = get_data_method(**kwargs)
+        all_data.extend(data.get('items'))
+        while nextp := data.get('next'):
+            data = get_data_method(page=nextp, **kwargs)
+            all_data.extend(data.get('items'))
+        return all_data
